#include <u.h>
#include <libc.h>
#include <bio.h>
#include "pedump.h"

typedef struct {
	char name[9];
	vlong vaddr;
	vlong base;
	vlong size;
} Secdir;

Biobuf *op;
static int Verbose;
static Secdir Sd[64];

static int
peekint(Pe *pe, int off, int width)
{
	int n;
	vlong was;

	was = Bseek(pe->bp, 0LL, 1);
	Bseek(pe->bp, off, 1);
	n = rint(pe, width);
	Bseek(pe->bp, was, 0);
	return n;
}

static vlong
peekmem(Pe *pe, int off, int len, char *buf)
{
	vlong n;
	vlong was;

	was = Bseek(pe->bp, 0LL, 1);
	Bseek(pe->bp, off, 1);
	n = Bread(pe->bp, buf, len);
	if(n == 0)
		sysfatal("peekmem: unexpected EOF\n");
	Bseek(pe->bp, was, 0);
	return n;
}

static vlong
getint(Pe *pe, int pos, int width)
{
	int n;
	vlong was;

	was = Bseek(pe->bp, 0LL, 1);
	Bseek(pe->bp, pos, 0);
	n = rint(pe, width);
	Bseek(pe->bp, was, 0);
	return n;
}

static void
getstr(Pe *pe, int pos, int len, char *buf)
{
	int c;
	vlong was;

	SET(c);
	was = Bseek(pe->bp, 0LL, 1);
	Bseek(pe->bp, pos, 0);
	while(len-- > 0 && (c = Bgetc(pe->bp)) > 0)
		*buf++ = c;
	*buf = 0;
	if(c < 0)
		sysfatal("peekmem: unexpected EOF\n");
	Bseek(pe->bp, was, 0);
}

void
dump_exports(Pe *pe, Secdir *sd)
{
	vlong ord, base, str, fun;
	char buf1[128], buf2[128];
	int i, name, nfunc, nname, funtab, namtab, ordtab;

	Bseek(pe->bp, sd->base, 0);
	rint(pe, 4);			/* flags (must be zero) */
	rint(pe, 4);			/* date stamp */
	rint(pe, 2);			/* major version */
	rint(pe, 2);			/* minor version */
	name = rint(pe, 4);
	base = rint(pe, 4);
	nfunc = rint(pe, 4);
	nname = rint(pe, 4);
	funtab = rint(pe, 4);
	namtab = rint(pe, 4);
	ordtab = rint(pe, 4);

	getstr(pe, sd->base+(name- sd->vaddr), sizeof(buf1), buf1);
	Bprint(op, "\t%s\n", buf1);

	for(i = 0; i < nname || i < nfunc; i++){
		fun = getint(pe, sd->base+(funtab- sd->vaddr)+(i*4), 4);
		ord = getint(pe, sd->base+(ordtab- sd->vaddr)+(i*4), 2);
		*buf1 = 0;
		if(i < nname){
			str = getint(pe, sd->base+(namtab- sd->vaddr)+(i*4), 4);
			getstr(pe, sd->base+(str- sd->vaddr), sizeof(buf1), buf1);
		}
		if(fun >= sd->vaddr && fun < sd->vaddr+sd->size){
			getstr(pe, sd->base+(fun- sd->vaddr), sizeof(buf2), buf2);
			Bprint(op, "\t\t%-6lld %s %s\n", ord+base, buf2, buf1);
		}
		else
			Bprint(op, "\t\t%-6lld 0x%-8llx %s\n", ord+base, fun+ sd->vaddr, buf1);
	}

}

void
dump_imports(Pe *pe, Secdir *sd)
{
	char buf[128];
	int orig, iat, hint, dll;

	Bseek(pe->bp, sd->base, 0);
	while(1){
		orig = rint(pe, 4);
		rint(pe, 4);			/* date stamp */
		rint(pe, 4);			/* forwarder chain */
		dll = rint(pe, 4);
		rint(pe, 4);			/* copy of orig unless app has been bound */
		if(dll == 0)
			break;
		getstr(pe, sd->base+(dll- sd->vaddr), sizeof(buf), buf);
		Bprint(op, "\t%s\n", buf);

		do{
			iat = getint(pe, sd->base+(orig- sd->vaddr), 4);
			if(iat == 0)
				break;
			if(iat & 0x80000000)
				Bprint(op, "\t\t%d\n", iat & 0x7fffffff);
			else{
				hint = getint(pe, sd->base+(iat- sd->vaddr), 2);
				getstr(pe, sd->base+(iat- sd->vaddr)+2, sizeof(buf), buf);
				Bprint(op, "\t\t%-6d %s\n", hint, buf);
			}
			orig += 4;
		}while(1);

	}
}

char *
getsect(int n)
{
	static char buf[32];

	switch(n){
	case -2:
		return "debug";
	case -1:
		return "abs";
	case 0:
		return "undef";
	default:
		break;
	}

	if(n > 0 && n <= nelem(Sd) && Sd[n].name != nil)
		return Sd[n].name;

	snprint(buf, sizeof(buf), "<%d>", n);
	return buf;
}

void
dump_symtab(Pe *pe, vlong symtab, int symcnt)
{
	uint *ip;
	vlong value, strtab;
	int i, j, idx, type, naux;
	char buf[128], *sect, *class, *base, *comp;

	strtab = symtab + symcnt * 0x12;
	
	naux = 0;
	Bseek(pe->bp, symtab, 0);
	for(i = 0; i < symcnt; i++){
		if(naux){
			for(j = 0; j < naux; j++)
				skip(pe, 0x12);
			naux = 0;
			continue;
		}

		rmem(pe, buf, 8);
		ip = (uint *)buf;
		if(ip[0] == 0){
			idx = ip[1];
			if(idx == 0)
				*buf = 0;
			else
				getstr(pe, strtab+idx, sizeof(buf), buf);
		}

		value = rint(pe, 4);
		sect = getsect((signed short)rint(pe, 2));
		type = rint(pe, 2);
		base = lookup(Symbase, type & 0xff);
		comp = lookup(Symcomp, (type >> 8) & 0xff);
		class = lookup(Symclass, rint(pe, 1));
		naux = rint(pe, 1);

		Bprint(op, "%-8s %-16s %-16s %-16s %-8lld %s\n", sect, base, comp, class, value, buf);

	}

	
}

void
headers(Pe *pe)
{
	Prfmt *f, *coffhdr, *nthdr;
	int i, petype, nsect, magic;
	int virtsz, rawsz, off, ndir, vaddr, size;
	int symtab, symcnt;

	/* DOS header */
	magic = getint(pe, 0, 2);
	if(magic != 0x5a4d)	// Mark Zbikowski, you are remembered
		sysfatal("0x%x  bad DOS header magic\n", magic);
	for(f = Doshdr; f->name; f++)
		pr(pe, f);

	off = rint(pe, 4);	/* Image header position in file */
	Bseek(pe->bp, off, 0);

	/* Image header */
	magic = rint(pe, 4);
	if(magic != 0x4550)
		sysfatal("0x%x: bad Image header magic\n", magic);
	Bprint(op, "\n");
	nsect = peekint(pe, 2, 2);
	symtab = peekint(pe, 8, 4);
	symcnt = peekint(pe, 12, 4);
	for(f = Imagehdr; f->name; f++)
		pr(pe, f);

	petype = peekint(pe, 0, 2);
	switch(petype){
	case 0x10b:
		coffhdr = Coffhdr32;
		nthdr = Nthdr32;
		break;
	case 0x20b:
		coffhdr = Coffhdr64;
		nthdr = Nthdr64;
		break;
	default:
		SET(coffhdr, nthdr);
		sysfatal("0x%x - unsupported file type\n", petype);
		break;
	}

	/* Coff header - first part of the mandatory "Optional" header */
	Bprint(op, "\n");
	for(f = coffhdr; f->name; f++)
		pr(pe, f);

	/* NT header - seccond part of the mandatory "Optional" header */
	Bprint(op, "\n");
	for(f = nthdr; f->name; f++)
		pr(pe, f);
	ndir = rint(pe, 4);

	/* data directory entries */
	Bprint(op, "\n");
	for(i = 0; i < ndir && i < 0x10; i++){
		vaddr = rint(pe, 4);
		size = rint(pe, 4);
		if(vaddr || size)
			Bprint(op, "%-32s     vaddr=%d size=%d\n", Direntries[i], vaddr, size);
	}

	Bprint(op, "\n");
	/* section headers */
	for(i = 0; i < nsect && i < nelem(Sd); i++){
		peekmem(pe, 0, 8, Sd[i].name);
		virtsz = peekint(pe, 16, 4);
		Sd[i].vaddr = peekint(pe, 12, 4);
		rawsz = peekint(pe, 16, 4);
		if(rawsz < virtsz)	/* apparantly this is what the win32 loader does */
			Sd[i].size = rawsz;
		else
			Sd[i].size = virtsz;
		Sd[i].base = peekint(pe, 20, 4);
		for(f = Sechdr; f->name; f++)
			pr(pe, f);
		Bprint(op, "\n");
	}

	/* section contents */
	for(i = 0; i < nsect && i < nelem(Sd); i++){
		Bprint(op, "section=%-8s addr=0x%x size=%d\n",
			Sd[i].name, Sd[i].vaddr, Sd[i].size);

		if(strcmp(Sd[i].name, ".idata") == 0)
			dump_imports(pe, &Sd[i]);
		else
		if(strcmp(Sd[i].name, ".edata") == 0)
			dump_exports(pe, &Sd[i]);
		else{
			Bseek(pe->bp, Sd[i].base, 0);
			if(Verbose)
				xd(pe, Sd[i].vaddr, Sd[i].size);
		}
		Bprint(op, "\n");
	}

	if(symtab && symcnt)
		dump_symtab(pe, symtab, symcnt);
}


void
dump(char *fn)
{
	Pe pe;

	pe.file = fn;
	if((pe.bp = Bopen(fn, OREAD)) == nil)
		sysfatal("%s - cannot open - %r\n", fn);
	headers(&pe);
	Bterm(pe.bp);
}

static void
usage(void)
{
	fprint(2, "usage: %s [-v] <exefile>...\n", argv0);
	exits("usage");
}

void
main(int argc, char *argv[])
{
	int i;
	Biobuf out;

	ARGBEGIN {
	case 'v':
		Verbose++;
		break;
	default:
		usage();
	} ARGEND

	Binit(&out, OWRITE, 1);
	op = &out;

	for(i = 0; i < argc; i++)
		dump(argv[i]);
	
	exits(nil);
}
